"""
Sub-module with melt-only thermometers
"""

import warnings as w

import elementMass as e
import pandas as pd

from MagmaPandas.MagmaFrames.protocols import Magma

from .data_parsing import _anhydrous_composition, _get_oxides, parse_data


def _check_calibration_range_Series(melt: pd.Series) -> None:

    SiO2_range = not (31 < melt["SiO2"] < 73.64)
    alkali_range = melt[["Na2O", "K2O"]].sum() > 14.3
    H2O_range = melt["H2O"] > 18.6

    outside_range = SiO2_range | alkali_range | H2O_range

    if not outside_range:
        return
    w.warn(f"sample has composition outside the thermometer calibration range")


def _check_calibration_range_dataframe(melt: Magma) -> None:

    SiO2_range = ~melt["SiO2"].between(31, 73.64)
    alkali_range = melt[["Na2O", "K2O"]].sum(axis=1) > 14.3
    H2O_range = melt["H2O"] > 18.6

    outside_range = SiO2_range | alkali_range | H2O_range

    if outside_range.sum() < 1:
        return

    w.warn(
        f"samples {*melt.index[outside_range],} have compositions outside the thermometer calibration range"
    )


def _check_calibration_range(melt: pd.Series | Magma) -> None:

    if isinstance(melt, pd.Series):
        _check_calibration_range_Series(melt)
    elif isinstance(melt, pd.DataFrame):
        _check_calibration_range_dataframe(melt)


def putirka2008_14(melt: Magma, *args, **kwargs) -> float | pd.Series:
    """
    melt-only thermometer

    Equation 14 from Putirka (2008)\ [15]_ calculates liquidus temperatures based on melt compositions.
    Requires equilibrium with olivine.

    SEE = 58 degrees

    Applicable between:

    -    0 - 14.4 GPa

    -    729 - 2000 degrees C

    -    31 - 73.64 wt. % SiO\ :sub:`2`

    -    0 - 14.3 wt. % Na\ :sub:`2`\ O + K\ :sub:`2`\ O

    -    0 - 18.6 wt. % H\ :sub:`2`\ O

    Parameters
    ----------
    melt : Magma
        melt compositions in oxide wt. %

    Returns
    -------
    temperatures : float, pandas Series
        liquidus temperatures in Kelvin.
    """

    _check_calibration_range(melt)

    oxides = _get_oxides(melt)
    oxides_needed = set(["MgO", "FeO", "Na2O", "K2O"])

    composition = parse_data(melt, oxides_needed)

    if "H2O" not in oxides:
        H2O = 0.0
    else:
        H2O = composition["H2O"].copy()
        composition = _anhydrous_composition(composition)

    # Calculate molar oxide fractions
    mol_fractions = composition.moles
    # Melt Mg#
    Mg_no = mol_fractions["MgO"] / (
        mol_fractions["MgO"] + mol_fractions["FeO"]
    )  # Putirka doesn't specify whether this should be Fe2+ or Fe(total)

    if "Fe2O3" in oxides:
        FeO_w, Fe2O3_w = e.compound_weights(["FeO", "Fe2O3"])
        composition["FeO"] = composition["FeO"] + (
            2 * composition["Fe2O3"] * (FeO_w / Fe2O3_w)
        )

    T_K = (
        754
        + 190.6 * Mg_no
        + 25.52 * composition["MgO"]
        + 9.585 * composition["FeO"]
        + 14.87 * (composition["Na2O"] + composition["K2O"])
        - 9.176 * H2O
    ) + 273.15

    return pd.Series(T_K, name="T_K").squeeze()


def putirka2008_15(
    melt: Magma, P_bar: float | pd.Series = None, **kwargs
) -> float | pd.Series:
    """
    melt-only thermometer

    Equation 15 from Putirka (2008)\ [15]_ calculates liquidus temperatures based on melt compositions.
    Requires equilibrium with olivine.

    SEE = 46 degrees

    Applicable between:

    -    0 - 14.4 GPa

    -    729 - 2000 degrees C

    -    31 - 73.64 wt. % SiO\ :sub:`2`

    -    0 - 14.3 wt. % Na\ :sub:`2`\ O + K\ :sub:`2`\ O

    -    0 - 18.6 wt. % H\ :sub:`2`\ O

    Parameters
    ----------
    melt : Magma
        melt compositions in oxide wt. %

    P_bar : float, pandas Series
        pressures in bar.

    Returns
    -------
    temperatures : float, pandas Series
        liquidus temperatures in Kelvin.
    """
    _check_calibration_range(melt)

    if P_bar is None:
        P_bar = melt["P_bar"]

    oxides = _get_oxides(melt)
    oxides_needed = set(["MgO", "FeO", "Na2O", "K2O"])
    composition = parse_data(melt, oxides_needed)

    if "H2O" not in oxides:
        H2O = 0.0
    else:
        H2O = composition["H2O"].copy()
        composition = _anhydrous_composition(composition)

    P_GPa = P_bar / 1e4

    # Calculate molar oxide fractions
    mol_fractions = composition.moles
    # Melt Mg#
    Mg_no = mol_fractions["MgO"] / (
        mol_fractions["MgO"] + mol_fractions["FeO"]
    )  # Putirka doesn't specify whether this should be Fe2+ or Fe(total)

    if "Fe2O3" in oxides:
        FeO_w, Fe2O3_w = e.compound_weights(["FeO", "Fe2O3"])
        composition["FeO"] = composition["FeO"] + (
            2 * composition["Fe2O3"] * (FeO_w / Fe2O3_w)
        )

    T_K = (
        815.3
        + 265.5 * Mg_no
        + 15.37 * composition["MgO"]
        + 8.61 * composition["FeO"]
        + 6.646 * (composition["Na2O"] + composition["K2O"])
        + 39.16 * P_GPa
        - 12.83 * H2O
    ) + 273.15

    return pd.Series(T_K, name="T_K").squeeze()


def putirka2008_16(
    melt: Magma, P_bar: float | pd.Series = None, **kwargs
) -> float | pd.Series:
    """
    melt-only thermometer

    Equation 16 from Putirka (2008)\ [15]_ calculates liquiqdus temperatures based on melt compositions.
    Requires equilibrium with olivine + plagioclase + clinopyroxene.

    SEE = 19 degrees

    Parameters
    ----------
    melt : Magma
        melt compositions in oxide wt. %

    P_bar : float, pandas Series
        pressures in bar

    Returns
    -------
    temperatures : float, pandas Series
        liquidus temperatures in Kelvin.
    """
    oxides = _get_oxides(melt)
    oxides_needed = set(["SiO2", "Al2O3", "MgO"])

    composition = parse_data(melt, oxides_needed)
    # import MagmaPandas as mp

    if P_bar is None:
        P_bar = melt["P_bar"]

    if isinstance(P_bar, pd.Series):
        if not melt.index.equals(P_bar.index):
            raise RuntimeError("Melt and P_bar indices don't match")

    if "H2O" in oxides:
        composition = _anhydrous_composition(composition)

    # Convert pressure from bars to GPa
    P_GPa = P_bar / 1e4

    # Calculate molar oxide fractions
    mol_fractions = composition.moles

    part_1 = (
        -583
        + 3141 * mol_fractions["SiO2"]
        + 15779 * mol_fractions["Al2O3"]
        + 1338.6 * mol_fractions["MgO"]
    )
    part_2 = -31440 * mol_fractions["SiO2"] * mol_fractions["Al2O3"] + 77.67 * P_GPa

    T_K = part_1 + part_2 + 273.15

    return pd.Series(T_K, name="T_K").squeeze()



from abc import ABC, abstractmethod


class Fe3Fe2_model(ABC):
    @abstractmethod
    def calculate_Fe3Fe2(
        cls, Melt_mol_fractions, T_K, fO2, *args, **kwargs
    ):
        pass

    @abstractmethod
    def get_error(cls, *args, **kwargs):
        pass

    @abstractmethod
    def get_offset_parameters(cls, n: int, *args, **kwargs):
        pass

    @abstractmethod
    def get_offset(cls, melt_composition, offset_parameters, *args, **kwargs):
        pass



from .Blundy import FeMg_blundy
from .Toplis import FeMg_Toplis

Kd_models = {"toplis": FeMg_Toplis, "blundy": FeMg_blundy}


def equilibrium_forsterite(Kd, Fe2Mg_liquid):
    """
    Parameters
    ----------
    Kd :
        (Fe_olivine / Fe_liquid) * (Mg_liquid / Mg_olivine) partitioning coefficient
    Fe2Mg
        melt Fe2+/Mg ratio

    Returns
    -------
    Equilibrium forsterite fraction as Mg/(Mg + Fe)
    """

    return 1 / (1 + Kd * Fe2Mg_liquid)



from functools import partial

import pandas as pd

from MagmaPandas import Fe_redox
from MagmaPandas.configuration import configuration
from MagmaPandas.fO2 import fO2_QFM
from MagmaPandas.MagmaFrames import MagmaFrame

from . import Kds_iterate as it
from . import Kds_vectorised as vec


def observed_FeMg_Kd(
    melt: MagmaFrame,
    forsterite: pd.Series,
    P_bar,
    T_K: None | float | pd.Series = None,
    **kwargs,
) -> pd.Series:
    """
    Calulate observed Kds based on given melt and olivine compositions.

    Parameters
    ----------
    melt : MagmaFrame
        melt composition in oxide wt. %
    forsterite : float, pandas Series
        olivine forsterite content
    P_bar   : float, pandas Series
        pressures in bar
    T_K : None, float, pandas Series
        temperatures in Kelvin. If set to None, temperatures are calculated according to the melt thermometer set in the global configuration.

    Returns
    -------
    Kds : pandas Series
    """
    if isinstance(P_bar, (int, float)):
        P_bar = pd.Series(P_bar, index=melt.index)
    if isinstance(forsterite, (int, float)):
        forsterite = pd.Series(forsterite, index=melt.index)

    forsterite = forsterite[melt.index]

    melt_x_moles = melt.moles

    Fe3Fe2_model = getattr(Fe_redox, configuration.Fe3Fe2_model)
    dQFM = kwargs.get("dQFM", configuration.dQFM)

    if T_K is None:
        T_K = melt_x_moles.convert_moles_wtPercent().temperature(P_bar)
    fO2 = fO2_QFM(dQFM, T_K, P_bar)
    Fe3Fe2 = Fe3Fe2_model(melt_x_moles, T_K, fO2)

    melt_x_moles = melt_x_moles.normalise()
    Fe2_FeTotal = 1 / (1 + Fe3Fe2)
    melt_MgFe = melt_x_moles["MgO"] / (melt_x_moles["FeO"] * Fe2_FeTotal)
    olivine_MgFe = forsterite / (1 - forsterite)
    Kd_observed = melt_MgFe / olivine_MgFe

    return Kd_observed


def calculate_FeMg_Kd(
    Melt_mol_fractions: pd.Series | pd.DataFrame,
    forsterite_initial: float | pd.Series,
    T_K: float | pd.Series,
    Fe3Fe2: float | pd.Series,
    *args,
    **kwargs,
) -> pd.Series:
    """
    Calculate equilibrium Kds for given melt compositions.

    Kd models are set in the global configuration

    Parameters
    ----------
    Melt_mol_fractions : pandas Series, pandas Dataframe
        melt composition in oxide mol fractions
    forsterite_initial  : float, pandas Series
        initial forsterite contents. Forsterite values are iteratively adjusted and initial values are not necessarily in Fe-Mg equilibrium with melts.
    T_K : float, pandas Series
        temperatures in Kelvin
    Fe3Fe2  : float, pandas Series
        melt |Fe3Fe2| ratios

    Returns
    -------
    Kds : pandas Series
    """

    if isinstance(Melt_mol_fractions, pd.Series):
        Kd_func = it.calculate_Kd
    elif isinstance(Melt_mol_fractions, pd.DataFrame):
        Kd_func = vec.calculate_Kd

    return Kd_func(
        Melt_mol_fractions=Melt_mol_fractions,
        forsterite_initial=forsterite_initial,
        T_K=T_K,
        Fe3Fe2=Fe3Fe2,
        *args,
        **kwargs,
    )



"""
Module with models for calculating melt Fe-Mg partition coefficients between olivine and melt as:

 (Fe\ :sup:`2+` / Mg)\ :sub:`ol` / (Fe\ :sup:`2+` / Mg)\ :sub:`melt`
"""

from .Kds import *
from .models import FeMg_blundy, FeMg_Toplis, Kd_models



import pandas as pd

from MagmaPandas.configuration import configuration
from MagmaPandas.parse_io import convert_to_series, match_indeces

from .models import Kd_models, equilibrium_forsterite


def calculate_Kd(
    Melt_mol_fractions: pd.DataFrame,
    forsterite_initial,
    T_K,
    Fe3Fe2,
    **kwargs,
):
    """

    Parameters
    ----------
    melt_mol_fractions : pd.DataFrame
        melt composition in oxide mol fraction.
    olivine_forsterite
        forsterite fraction in olivine as Mg * 100 / (Mg + Fe)
    T_K :
        Temperature in Kelvin
    P_bar :
        Pressure in bar
    Fe3Fe2 :
        melt Fe3+/Fe2+ ratio
    """

    model_name = kwargs.pop("model", configuration.Kd_model)
    Kd_model = Kd_models[model_name]

    match_indeces(Melt_mol_fractions, [T_K, Fe3Fe2, *kwargs.values()])

    melt_mol_fractions = Melt_mol_fractions.copy()

    forsterite, T_K, Fe3Fe2, args = convert_to_series(
        [forsterite_initial, T_K, Fe3Fe2, *kwargs.values()], melt_mol_fractions.index
    )
    if len(kwargs.keys()) < 2:
        args = [args]

    kwargs = {key: value for key, value in zip(kwargs.keys(), args)}

    fo_converge_default = 0.001
    fo_converge = kwargs.pop("fo_converge", fo_converge_default)

    # initialise Kds
    Kd = Kd_model.calculate_Kd(melt_mol_fractions, forsterite, T_K, **kwargs)

    Fe2FeTotal = 1 / (1 + Fe3Fe2)
    Fe2Mg = melt_mol_fractions["FeO"] * Fe2FeTotal / melt_mol_fractions["MgO"]

    forsterite_EQ = equilibrium_forsterite(Kd, Fe2Mg)
    # Difference between observed Fo and equilibrium Fo
    forsterite_delta = abs(forsterite - forsterite_EQ) / forsterite

    iterate = forsterite_delta > fo_converge
    # iterate until equilibrium forsterite content doesn't change any more
    while sum(iterate) > 0:

        Kd[iterate] = Kd_model.calculate_Kd(
            melt_mol_fractions.loc[iterate],
            forsterite_EQ.loc[iterate],
            T_K[iterate],
            **{key: value[iterate] for key, value in kwargs.items()},
        )
        forsterite[iterate] = forsterite_EQ[iterate].copy()
        forsterite_EQ[iterate] = forsterite_EQ = equilibrium_forsterite(
            Kd[iterate], Fe2Mg[iterate]
        )

        forsterite_delta[iterate] = (
            abs(forsterite[iterate] - forsterite_EQ[iterate]) / forsterite[iterate]
        )
        iterate = forsterite_delta > fo_converge

    return Kd



"""
Sub-module with olivine-melt thermometers
"""

import numpy as np
import pandas as pd

from MagmaPandas.MagmaFrames.protocols import Magma


def putirka2007_4(
    liquid: Magma, olivine: Magma, P_bar: float | pd.Series, **kwargs
) -> pd.Series:
    """
    Olivine-liquid thermometer

    Equation 4 from Putirka (2007)\ [16]_ calculates liquidus temperatures based on olivine-melt pairs.
    Identical to equation 22 from Putirka (2008)\ [15]_.

    SEE = hydrous: 29, anhydrous: 45, total: 43

    Total Fe is expressed as FeO

    Parameters
    ----------
    liquid  : Magma
        melt compositions in oxide wt. %
    olivine : pandas Series, pandas Dataframe
        olivine compositions in oxide wt. %
    P_bar : float, pandas Series
        pressures in bar

    Returns
    -------
    temperatures : pd.Series
        olivine liquidus temperatures in Kelvin.
    """
    from ..MagmaFrames import MagmaFrame
    from ..MagmaSeries import MagmaSeries

    composition = liquid.copy(deep=True)
    composition = composition.fillna(0.0)

    if isinstance(composition, MagmaFrame):
        elements = composition.columns
    elif isinstance(composition, MagmaSeries):
        elements = composition.index

    oxides = set(["MgO", "FeO", "Na2O", "K2O", "CaO", "SiO2", "K2O", "TiO2"])
    optional_oxides = {"MnO", "CoO", "NiO"}
    absentOxides = oxides.difference(elements)
    fill_oxides = optional_oxides.difference(elements)

    if len(absentOxides) > 0:
        raise KeyError(f"{absentOxides} not found in melt")

    # All oxides as mole/cation fractions on an anhydrous basis
    if len(fill_oxides) > 0:
        composition[list(fill_oxides)] = 0.0

    if "H2O" not in elements:
        H2O = 0.0
    else:
        H2O = composition["H2O"].copy()
        # moles are calculated on an anhydrous basis
        try:
            composition = composition.drop("H2O")
        except KeyError:
            composition = composition.drop(columns=["H2O"])

    # weight % oxides not renormalised to 100%, even when hydrous
    composition = composition.recalculate()

    # Calculate molar oxide fractions
    liquid_moles = composition.moles
    liquid_cations = composition.cations
    olivine_cations = olivine.cations

    C_NM = liquid_cations[["Fe", "Mn", "Mg", "Ca", "Co", "Ni"]].sum(
        axis=1
    )  # Fe = Fe total
    NF = 7 / 2 * np.log(1 - liquid_cations["Al"]) + 7 * np.log(1 - liquid_cations["Ti"])
    D_Mg = olivine_cations.loc[liquid_cations.index, "Mg"] / liquid_cations["Mg"]

    P_GPa = P_bar / 1e4

    numerator = 15294.6 + 1318.8 * P_GPa + 2.4834 * (P_GPa**2)
    denominator_a = 8.048 + 2.8532 * np.log(D_Mg) + 2.097 * np.log(1.5 * C_NM)
    denominator_b = (
        2.575 * np.log(3 * liquid_moles["SiO2"]) - 1.41 * NF + 0.222 * H2O + 0.5 * P_GPa
    )

    denominator = denominator_a + denominator_b

    return (numerator / denominator) + 273.15



from importlib import resources

import matplotlib.pyplot as plt
import pandas as pd


def TAS(ax, labels=False, fontsize="medium", **kwargs):
    """Returns a line plot element of classification of volcanic rocks
    in total-alkali vs silica plots
    """

    # resources.open_text("geoplot.data", "TAS.csv")
    with resources.files("geoplot.data").joinpath("TAS.csv").open("r") as df:
        TAS = pd.read_csv(df)

    rock_labels = {
        "Picro-basalt": ["Picro\nbasalt", [41.7, 1.5]],
        "Basalt": ["Basalt", [47, 2.5]],
        "Basaltic andesite": ["Basaltic\nandesite", [53, 2.5]],
        "Andesite": ["Andesite", [58, 2.5]],
        "Dacite": ["Dacite", [65.5, 4]],
        "Trachy-basalt": ["Trachy-\nbasalt", [47.5, 5.5]],
        "Basaltic trachy-andesite": ["Basaltic\ntrachy-\nandesite", [51.6, 6.5]],
        "Trachy-andesite": ["Trachy-\nandesite", [56, 8]],
        "Trachyte": ["Trachyte", [64, 11]],
        "Tephrite": ["Tephrite", [43.5, 7]],
        "Phono-tephrite": ["Phono-\ntephrite", [47, 9.0]],
        "Tephri-phonolite": ["Tephri-\nphonolite", [51, 11]],
        "Phonolite": ["Phonolite", [55, 15]],
        "Foidite": ["Foidite", [45, 14]],
        "Rhyolite": ["Rhyolite", [72, 8.5]],
    }

    if labels:
        for _, rock in rock_labels.items():
            ax.text(
                *rock[1],
                rock[0],
                fontsize=fontsize,
                fontfamily="monospace",
                clip_on=True,
            )

    for id in TAS.id.unique():
        ax.plot(
            TAS.loc[TAS.id == id, "x"],
            TAS.loc[TAS.id == id, "y"],
            "-",
            color="k",
            **kwargs,
        )



import numpy as np
import pandas as pd

from MagmaPandas import Fe_redox, configuration
from MagmaPandas.fO2 import fO2_QFM

from ...Kd_baseclass import Kd_model


class FeMg_blundy(Kd_model):
    """
    Calculate equilibrium Fe-Mg partition coefficients between olivine and melt according to equation 8 from Blundy (2020)\ [11]_.
    """

    errors = pd.Series({6: 0.019, 9: 0.04, 100: 0.063})

    @classmethod
    def _get_Fe3FeTotal(Melt_mol_fractions, T_K, P_bar=1, **kwargs):
        Fe3Fe2_model = Fe_redox.Fe3Fe2_borisov
        dQFM = kwargs.get("dQFM", configuration.dQFM)

        fO2 = fO2_QFM(dQFM, T_K, P_bar)
        Fe3Fe2 = Fe3Fe2_model.calculate_Fe3Fe2(Melt_mol_fractions, T_K, fO2)

        return Fe3Fe2 / (1 + Fe3Fe2)

    @classmethod
    def calculate_Kd(
        cls, Melt_mol_fractions: pd.DataFrame, forsterite, T_K, P_bar=1, *args, **kwargs
    ) -> float | pd.Series:
        """
        calculate equilibrium Kds for given melt compositions.

        Parameters
        ----------
        Melt_mol_fractions  : pandas Dataframe
            melt composition in oxide mol fractions
        forsterite  : float, array-like
            initial olivine forsterite content. Forsterite values are iteratively adjusted and initial values are not necessarily in Fe-Mg equilibrium with melts.
        T_K : float, array-like
            temperatures in Kelvin
        P_bar   : float array-like
            pressures in bar

        Returns
        -------
        Kds : float, array-like
        """

        Fe3FeTotal = cls._get_Fe3FeTotal(Melt_mol_fractions, T_K, P_bar)

        return 0.3642 * (1 - Fe3FeTotal) * np.exp(312.7 * (1 - 2 * forsterite) / T_K)

    @classmethod
    def get_error(cls, melt_composition: pd.DataFrame) -> float | pd.Series:
        """
        Calculate one standard deviation errors on Kds

        Parameters
        ----------
        melt_composition    : pandas Dataframe
            melt composition in oxide wt. %

        Returns
        -------
        error:  float, array-like
            one standard deviation error
        """

        axis = [0, 1][isinstance(melt_composition, pd.DataFrame)]
        alkalis = melt_composition[["Na2O", "K2O"]].sum(axis=axis)

        if isinstance(alkalis, (int, float)):

            composition_filter = np.less(alkalis, cls.errors.index.values)
            idx = cls.errors.index.values[composition_filter][0]

            return cls.errors[idx]

        errors = pd.Series(index=melt_composition.index)

        for i, a in alkalis.items():
            composition_filter = np.less(a, cls.errors.index.values)
            idx = cls.errors.index.values[composition_filter][0]
            errors.loc[i] = cls.errors[idx]

        return errors

    @classmethod
    def get_offset_parameters(cls, n=1):
        return np.random.normal(loc=0, scale=1, size=n)

    @classmethod
    def get_offset(cls, melt_composition, offset_parameters, *args, **kwargs):

        error = cls.get_error(melt_composition=melt_composition)

        return offset_parameters * error



import inspect
from functools import wraps
from typing import List


def _check_argument(var_name: str, allowed_values: List[str]):
    def decorator(func):
        """
        Check if var_name has a valid value
        """
        # Get values for default arguments
        sig = inspect.signature(func)
        defaults = {
            n: val
            for n, val in zip(
                sig.parameters.keys(), [val.default for val in sig.parameters.values()]
            )
        }

        @wraps(func)  # comment
        def wrapper(*args, **kwargs):

            for key in kwargs.keys():
                # Remove user kwargs from the default dict
                defaults.pop(key, None)

            var = {**kwargs, **defaults}.get(var_name, None)
            if var not in allowed_values:
                raise ValueError(
                    f"{var_name}: {var}, not recognised, please choose from: {allowed_values}"
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def _check_setter(allowed_values):
    def decorator(func):
        """
        check if set value is valid
        """

        @wraps(func)  # comment
        def wrapper(*args):
            var = args[1]
            if isinstance(var, (float, int)):
                min, max = allowed_values
                if not (min < var < max):
                    raise ValueError(
                        f"value: {var}, not within bounds {*allowed_values,}"
                    )
            elif isinstance(var, str):
                if var not in allowed_values:
                    raise ValueError(
                        f"'{var}' is not recognised, please choose from: {*allowed_values,}"
                    )
            return func(*args)

        return wrapper

    return decorator


def _check_attribute(attr_name: str, allowed_values: List[str]):
    def decorator(func):
        """
        Check if attr_name has a valid value
        """

        @wraps(func)  # comment
        def wrapper(self, *args, **kwargs):
            attr = getattr(self, attr_name)
            if attr.value not in allowed_values:
                raise ValueError(
                    f"Calculation is not valid with: {self.units}, please use: {*allowed_values,}"
                )
            return func(self, *args, **kwargs)

        # wrapper.__doc__ = func.__doc__
        return wrapper

    return decorator


def _check_value(var_name: str, allowed_range: List[float]):
    def decorator(func):
        """
        Check if var_name has a valid value
        """

        @wraps(func)  # comment
        def wrapper(*args, **kwargs):
            var = kwargs.get(var_name, None)
            min, max = allowed_range
            if min < var < max:
                raise ValueError(
                    f"{var_name}: {var}, outside allowed range: {min} - {max}"
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator



def _check_oxides(composition, oxides_needed):

    oxides = _get_oxides(composition)

    absentOxides = oxides_needed.difference(oxides)

    if len(absentOxides) > 0:
        raise KeyError(f"{absentOxides} not found in melt")


def _fill_data(composition):

    composition_filled = composition.copy(deep=True)
    composition_filled = composition_filled.fillna(0.0)

    return composition_filled


def _get_oxides(composition):

    import MagmaPandas as mp

    if isinstance(composition, mp.MagmaFrame):
        return composition.columns

    if isinstance(composition, mp.MagmaSeries):
        return composition.index


def parse_data(composition, oxides_needed):

    _check_oxides(composition, oxides_needed)

    return _fill_data(composition)


def _anhydrous_composition(composition):

    composition_H2O = composition.copy()

    try:
        composition_H2O = composition_H2O.drop("H2O")
    except KeyError:
        composition_H2O = composition_H2O.drop(columns=["H2O"])

    return composition_H2O.recalculate()



"""
Module with melt-only and melt-mineral thermometers
"""

import inspect

from . import melt, ol_melt

melt_thermometers = {
    func[0]: func[1] for func in inspect.getmembers(melt, inspect.isfunction)
}

ol_melt_thermometers = {
    func[0]: func[1] for func in inspect.getmembers(ol_melt, inspect.isfunction)
}



import numpy as np
import pandas as pd

from ..Fe_redox_baseclass import Fe3Fe2_model


class Fe3Fe2_kressCarmichael(Fe3Fe2_model):
    """
    Calculate melt |Fe3Fe2| ratios according to equation 7 from Kress and Carmichael (1991)\ [2]_.
    """

    FeO_error = 0.21
    Fe2O3_error = 0.42

    @classmethod
    def calculate_Fe3Fe2(cls, Melt_mol_fractions, T_K, fO2, P_bar):
        """
        Calculate melt |Fe3Fe2| ratios\.

        Parameters
        ----------
        mol_fractions   :  :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
            melt composition in oxide mol fractions
        T_K        :     float, array-like
            temperature in Kelvin
        fO2     :      float, array-like
            Oxygen fugacity
        P_Pa      :      float, array-like
            Pressure in Pascals


        Returns
        -------
        float, array-like
            melt |Fe3Fe2| ratio
        """
        P_Pa = P_bar * 1e-5

        LNfO2 = np.log(fO2)

        components = ["Al2O3", "FeO", "CaO", "Na2O", "K2O"]

        # Parameters from table 7
        a = 0.196
        b = 1.1492e4
        c = -6.675

        dCoefficients = pd.Series(
            {"Al2O3": -2.243, "FeO": -1.828, "CaO": 3.201, "Na2O": 5.854, "K2O": 6.215},
            name="dCoeff",
        )

        e = -3.36
        f = -7.01e-7
        g = -1.54e-10
        h = 3.85e-17
        T0 = 1673
        sumComponents = (
            Melt_mol_fractions.loc[:, components].mul(dCoefficients).sum(axis=1)
        )

        part1 = a * LNfO2 + b / T_K + c + sumComponents
        part2 = e * (1 - T0 / T_K - np.log(T_K / T0))
        part3 = f * P_Pa / T_K + g * ((T_K - T0) * P_Pa) / T_K + h * P_Pa**2 / T_K

        return 2 * np.exp(part1 + part2 + part3)

    @classmethod
    def get_error(cls, melt_composition, Fe3Fe2):
        """
        Calculate one standard deviation errors on |Fe3Fe2| ratios, based on reported calibration errors of 0.21 and 0.42 wt.% on FeO and |Fe2O3|.

        Parameters
        ----------
        melt_composition : :py:class:`~MagmaPandas.MagmaFrames.magmaFrame.MagmaFrame`
            melt composition in wt. % oxides
        Fe3Fe2  :   float, array-like
            melt |Fe3Fe2| ratios

        Returns
        -------
        float, array-like
            |Fe3Fe2| error
        """

        melt = melt_composition.FeO_Fe2O3_calc(Fe3Fe2, total_Fe="FeO", inplace=False)

        Fe2O3_contribution = cls.FeO_error / 2 / melt["Fe2O3"]
        FeO_contribution = cls.FeO_error / melt["FeO"]

        return (Fe2O3_contribution + FeO_contribution) * Fe3Fe2

    @classmethod
    def get_offset_parameters(cls, n: int = 1):
        return np.random.normal(loc=0, scale=1, size=n)

    @classmethod
    def get_offset(cls, melt_composition, Fe3Fe2, offset_parameters, *args, **kwargs):
        FeO_offset = offset_parameters * cls.FeO_error
        Fe2O3_offset = (
            FeO_offset * 2
        )  # Are FeO and Fe2O3 errors really correlated this way??

        melt = melt_composition.FeO_Fe2O3_calc(Fe3Fe2, total_Fe="FeO", inplace=False)

        Fe2O3_contribution = Fe2O3_offset / 2 / melt["Fe2O3"]
        FeO_contribution = FeO_offset / melt["FeO"]

        return (Fe2O3_contribution + FeO_contribution) * Fe3Fe2



from typing import List, Union

import pandas as pd


def match_indeces(data: Union[pd.DataFrame, pd.Series], variables: List[pd.Series]):

    for var in variables:

        try:
            if not data.index.equals(var.index):
                raise RuntimeError(f"Data and {var.name} indices don't match")
        except AttributeError:
            pass


def convert_to_series(variables: List, index):

    new_vars = []
    for var in variables:
        if isinstance(var, (int, float)):
            new_var = pd.Series(var, index=index)
        else:
            new_var = var.copy()
        new_vars.append(new_var)

    return new_vars
